package edu.toronto.cs.ctl;

import java.util.*;
import java.io.*;
import edu.toronto.cs.tlq.*;
import edu.toronto.cs.algebra.*;
import edu.toronto.cs.util.caching.*;

/**
 ** This class does some macro expansion on CTL parse trees. It also can
 ** perform some macro compression.
 **/
public class CTLMacroManager
{
  // -- map between macro name -- macro expansion
  Cache macros;
  // -- inverse of the macros map
  Cache expansions;

  //PlaceholderManager phman;

  List macroExpanders;

  /**
   ** Constructs a new CTL macro manager.
   **/
  public CTLMacroManager (FiniteLattice lat)
  {
    macros = new UnboundedMapCache ();
    expansions = new UnboundedMapCache ();
    macroExpanders = new ArrayList ();

//     if (lat instanceof UpSetLattice)
//       phman = new PlaceholderManager ((UpSetLattice) lat);
//     else phman = null; // no placeholder manager...
  }

  // -- resets the macro manager 
  public void reset ()
  {
    macros = new UnboundedMapCache ();
    expansions = new UnboundedMapCache ();
//     if (phman != null) 
//       phman = new PlaceholderManager (phman.getLattice ());

    for (Iterator it = macroExpanders.iterator (); it.hasNext ();)
      ((MacroExpander)it.next ()).reset ();
    
  }

  public void addExpander (MacroExpander m)
  {
    macroExpanders.add (m);
    m.setManager (this);
  }
  
  public void prependExpander (MacroExpander m)
  {
    if (macroExpanders.isEmpty ()) 
      macroExpanders.add (m);
    else
      macroExpanders.add (0, m);
    m.setManager (this);
  }
  
  /**
   ** Defines a new macro. If a macro with such a name already exists
   ** it gets overwritten. (There is no support for defining macros
   ** containing other macros, i.e. it may or may not work)
   **/
  public void defineMacro (CTLNode name, CTLNode expansion)
  {
    // using null as a macro is not allowed
    if (name == null || expansion == null) return;
    // for expanding macros
    macros.put (name, expansion);
    // for compressing macros
    expansions.put (expansion, name);
  }

  /**
   ** Remove the specified macro.
   **/
  public void removeMacro (CTLNode name)
  {
    // using null as a macro is not allowed
    if (name == null) return;
    Object value = macros.get (name);
    macros.remove (name);
    expansions.remove (value);
  }

  /**
   ** Expands all the macros present in the CTL sub-tree.
   **
   ** @return -- CTL parse tree WITHOUT marcoes
   **/
  public CTLNode expandMacro (CTLNode ctl)
  {
//     if (macros.isEmpty ()) return ctl; // no macros defined
//     else return expand (ctl, macros);
    // -- potentially there are no macros but there are placeholders
    // -- and thus we must proceed with expansion
    return expand (ctl, macros);
  }
  
  /**
   ** Compresses all the macros it can find in the CTL sub-tree.
   **
   ** @return -- CTL parse tree WITH macros (if any were found)
   **/
  public CTLNode compressMacro (CTLNode ctl)
  {
    if (expansions.isEmpty ()) return ctl; // can't compress
    else return expand (ctl, expansions); // no need for the lattice
  }

  /**
   ** Get a list of all the macro names.
   **/
  public Collection getAllMacroNames ()
  {
    return expansions.values ();
  }

  CTLNode expand (CTLNode ctl, Cache macroCache)
  {
    // -- CTL tree may have null children
    if (ctl == null) return null;
    
    // -- check if this was handled already
    if (macroCache.containsKey (ctl))
      return (CTLNode) macroCache.get (ctl);
    
    // -- try and feed this CTL to all expanders one after another
    CTLNode newCTL = ctl;
    
    for (Iterator it = macroExpanders.iterator (); it.hasNext ();)
      {
	MacroExpander expander = (MacroExpander)it.next ();
	newCTL = expander.expand (newCTL);
      }
    
    // -- check if anything was changed
    if (newCTL != ctl)
      {
	defineMacro (ctl, newCTL);
	return newCTL;
      }
    
    // -- if we got here then no expansion took place so continue traversing
    // -- the tree
    return CTLFactory.alterChildren (ctl, 
				     expand (ctl.getLeft (), macroCache),
				     expand (ctl.getRight (), macroCache));
  }
  

  /**
   ** The core macro expander.
   **/
//   private CTLNode _expand (CTLNode ctl, Cache expander)
//   {
//     // not much to expand here...
//     if (ctl == null) return null;
    
//     // found something to expand
//     if (expander.containsKey (ctl))
//       return (CTLNode) expander.get (ctl);

//     // found a placeholder - create expansion
//     else if (ctl instanceof CTLPlaceholderNode)
//       {
// 	CTLNode exp = phman.expandPlaceholder ((CTLPlaceholderNode) ctl);
// 	defineMacro ((CTLPlaceholderNode) ctl, exp);
// 	return exp;
//       }
//     else // expand the node's children
//       {
// 	// kind of a HACK: if the node has no children, it will not be
// 	// actually altered (no new node will be created)
// 	return CTLFactory.alterChildren (ctl,
// 					 (CTLNode) expand (ctl.getLeft (),
// 							   expander),
// 					 (CTLNode) expand (ctl.getRight (),
// 							   expander));
//       } 
//   }
  
}
