package edu.toronto.cs.ctl;

import java.util.*;
import java.io.*;
import edu.toronto.cs.algebra.*;
import edu.toronto.cs.tlq.*;
import edu.toronto.cs.util.caching.*;
import edu.toronto.cs.smv.*;

/**
 ** This class allows to convert CTLPlaceholderNode into something
 ** that can be used by the model checker and CTLMacroManager.
 **/
public class PlaceholderManager2 extends PlaceholderManager
{

  // -- list of term expansions for each placeholder
  // -- each element is of type CTLNode[]
  List placeHolderTerms;
  
  // -- name of all place holders that we know about
  // -- the expanded term for each placeholder are in 'terms', 
  // -- i.e. terms.get (i) correspond to placeHolders.get (i)
  List placeHolders;
  
  
  // -- the top free bit
  int topBit;
  
  
  UpSetLattice lat;

  EnumeratedTypes enumTypes;

  /**
   ** Constructs a new placeholder manager.
   **/
  public PlaceholderManager2 (UpSetLattice lattice, EnumeratedTypes _enumTypes)
  {
    placeHolderTerms = new ArrayList ();
    placeHolders = new ArrayList ();
    
    lat = lattice;
    enumTypes = _enumTypes;
    if (enumTypes == null)
      enumTypes = EnumeratedTypes.EMPTY_TYPES;
    topBit = 0;
    
  }

  public PlaceholderManager2 (UpSetLattice lattice)
  {
    this (lattice, EnumeratedTypes.EMPTY_TYPES);
  }
  

  public void reset ()
  {
    placeHolderTerms = new ArrayList ();
    placeHolders = new ArrayList ();
    topBit = 0;
  }
  
  public CTLNode expand (CTLNode ctl)
  {
    if (ctl instanceof CTLPlaceholderNode)
      return expandPlaceholder ((CTLPlaceholderNode)ctl);
    
    return ctl;
  }
  

  public UpSetLattice getLattice ()
  {
    return lat;
  }
  

  public CTLNode expandPlaceholder (CTLPlaceholderNode ph)
  {
    // -- add the place holder to the list of known placeholders
    placeHolders.add (ph);
    
    // -- get a list of atomic propositions
    List props = Arrays.asList (ph.getProps ());
    

    // -- build all propositions
    List terms = buildTerms (props.iterator ());
    
    CTLNode result = null;
    
    // -- allocate a lattice value for each term and build 
    // -- a resulting CTL node

    for (Iterator it = terms.iterator (); it.hasNext (); )
      {
	// -- get a minterm
	CTLNode minterm = (CTLNode)it.next ();

	// -- create a lattice node
	CTLNode latNode =
	  CTLFactory.createCTLLatticeNode (lat.getUpMinTerm (topBit++), lat);

	// -- add to the result
	if (result == null)
	  result = minterm.and (latNode);
	else
	  result = result.or (minterm.and (latNode));
      }
    
    // -- add terms to the list
    placeHolderTerms.add (terms.toArray (new CTLNode [terms.size ()]));
    
    // -- we are all done
    return result;
    
  }
  
  // -- recursively builds minterms
  List buildTerms (Iterator props)
  {
    String prop = (String)props.next ();
    
    // -- if there are no more props to look at
    if (!props.hasNext ())
      return buildTermForProp (prop);
    
    // -- build a list for this prop
    List propTerms = buildTermForProp (prop);
    
    // -- build a list for remaining propositions
    List theRest = buildTerms (props);
    
    // -- merge the two lists
    List result = new ArrayList ();
    
    for (Iterator it = propTerms.iterator (); it.hasNext ();)
      {
	CTLNode propNode = (CTLNode)it.next ();
	for (Iterator jt = theRest.iterator (); jt.hasNext ();)
	  result.add (propNode.and ((CTLNode)jt.next ()));
      }
 
    // -- all done 
    return result;
  }
  
  List buildTermForProp (String prop)
  {

    if (enumTypes.isEnumeratedVariable (prop))
      return buildEnumTerms (prop);
    
    // -- not enumerated type so we handle it ourselves
    List list = new ArrayList ();
    CTLNode propNode = CTLFactory.createCTLAtomNode (prop);
    list.add (propNode);
    list.add (propNode.neg ());
    return list;
  }
  

  List buildEnumTerms (String prop)
  {
    // -- get the range of this variable
    List range = enumTypes.getRange (prop);
    
    CTLNode var = CTLFactory.createCTLAtomNode (prop);
    
    List result = new ArrayList ();
    
    for (Iterator it = range.iterator (); it.hasNext (); )
      result.add (var.eq (CTLFactory.createCTLAtomNode (it.next ())));
    
    return result;
  }
  

  // -- given a lattice value interpret it using propositional formulas
  public List interpretSoln (int latVal)
  {
    List solution;
    if (latVal == lat.idtimesIdx () || latVal == lat.idplusIdx ())
      {
	solution = new LinkedList ();
	solution.add (CTLFactory.createCTLLatticeNode (latVal, lat));
      }
    else
      solution = interpretSoln ((int[]) lat.getElementKeyByIdx (latVal));

    return solution;
  }
  
  /**
   ** Given an array of integers returns their interpretation as a
   ** List of solutions.
   **/
  public List interpretSoln (int [] soln)
  {
    List solution = new LinkedList ();
    
    for (int i = 0; i < soln.length; i++)
      solution.add (intToSoln (soln [i]));

    return solution;
  }



  CTLNode intToSoln (int sol)
  {
    int length = placeHolderTerms.size ();
    CTLNode result = null;

    for (int i = 0; i < length; i++)
      {
	
	CTLNode[] terms = (CTLNode[])placeHolderTerms.get (i);
	CTLNode tempResult = (CTLNode)placeHolders.get (i);
	tempResult = tempResult.eq (intToSoln (sol, terms));
	
	if (result == null)
	  result = tempResult;
	else
	  result = result.and (tempResult);

	// -- removed bits already processed
	sol = sol >>> terms.length;
      }
    return result;
  }
  
  CTLNode intToSoln (int sol, CTLNode[] terms)
  {
    CTLNode solution = null;
    int bits = sol;
    
    // -- the bit we are curently considering
    for (int nbit = 0; bits != 0 && nbit < terms.length; nbit++)
      {
	if ((bits & 1) == 1)
	  {
	    if (solution == null) 
	      solution = terms [nbit];
	    else 
	      solution = solution.or (terms [nbit]);
	  }

	// -- remove the bit we've processed
	bits = bits >>> 1;
      }

    // -- it is possible that solution is still null, if 
    // -- for this place holder anything (including F) is 
    // -- a solution, so we hack it up a bit to see the desired
    // -- output
    if (solution == null)
      solution =
	CTLFactory.createCTLLatticeNode (lat.idtimesIdx (), lat);
    return solution;
  }  
}


