package edu.toronto.cs.cudd;
import edu.toronto.cs.util.Logger;
import edu.toronto.cs.util.Assert;
import edu.toronto.cs.algebra.*;

/***
 *** Wrapper for CUDD to allow the use of ADDs with our lattices
 ***/
public class CuddLatAdd
{
  static Logger log = new Logger (CuddLatAdd.class.getName ());
  
  // -- pointer to the CuddLatAdd manager
  int mgrPtr;

  ADD addOne = null;
  ADD addZero = null;

  static 
  {
    System.loadLibrary ("jcudd");
  }


  // -- if hasTables is true we can obtain the meet and join tables
  // -- otherwise we must use the lattice object to perform operations
  // -- on the terminals
  public CuddLatAdd (Algebra lattice, int initVars)
  {
    /***
     *** Currently the ADD library supports two modes of operation
     *** If the lattice contains meet/join/neg tables they are just
     *** copied over into C memory space and all operations on the terminal
     *** nodes are performed using these tables there 
     ***
     *** Alternatively if such tables are not available, for example they 
     *** would be to large to construct, the actual Algebra object is passed
     *** to the C land, and operations on terminal nodes are resolved through
     *** it. Potentially this means that each operation on a terminal nodes
     *** requires a callback which may be fairly expensive. However, the 
     *** alternative is two implement the lattice in C land does not look
     *** any better at the moment!
     ***/
    log.debug ("Initializing");

    // XXX removed for testing 
//     if (lattice instanceof FiniteLattice)
//       {
// 	FiniteLattice lat = (FiniteLattice)lattice;
// 	init (initVars,
// 	      lat.idtimesIdx (), 
// 	      lat.idplusIdx (),
// 	      lat.getMeetTable (),
// 	      lat.getJoinTable (),
// 	      lat.getNegTable ());
//       }
//     else
      init (lattice, initVars, lattice.idtimesIdx (), lattice.idplusIdx ());
  }
  
  public CuddLatAdd (Algebra lattice)
  {
    this (lattice, 0);
  }



  native void init (Algebra lattice, int intiVars, int top, int bot);
  
  // -- initialize the mgrPtr
  native void init (int initVars, int top, int bot, int[][] meetTable, 
		    int[][] joinTable, int[] negTable);

  private native int cuddAddOnePtr ();
  private native int cuddAddZeroPtr ();

  public int addOnePtr ()
  {
    //return addOne ().getNodePtr ();
    return cuddAddOnePtr ();
  }
  public int addZeroPtr ()
  {
    //return addZero ().getNodePtr ();
    return cuddAddZeroPtr ();
  }
  
  

  public native synchronized int addVarPtr (int index);  
  public native synchronized int addConstant (int value);

  public native synchronized int addAndPtr (int a, int b);
  public native synchronized int addOrPtr (int a, int b);
  public native synchronized int addNotPtr (int a);

  public native synchronized int addGeqPtr (int a, int b);
  public native synchronized int addLeqPtr (int a, int b);
  public native synchronized int addEqPtr (int a, int b);

  public native synchronized int addConstrainPtr (int f, int c);
  public native synchronized int addPermutePtr (int a, int[] map);
  public native synchronized int addGetValue (int a);

  public native synchronized int existAbstract (int ptr, int cube);
  public native synchronized int forallAbstract (int ptr, int cube);
  public native synchronized void ref (int ptr);
  public native synchronized void deref (int ptr);
  public native synchronized void recursiveDeref (int ptr);

  
  

  public native synchronized int checkZeroRef ();
  public native synchronized void quit ();

  // -- outputs info on stdout
  public native synchronized void info ();

  public native synchronized void printMintermPtr (int a);

  public native synchronized void reorder (int minsize);
  public native synchronized int dagSize (int ptr);
  public native synchronized int sharingSize (int[] ptrs);


  public native synchronized int dumpDaVinciPtr (String fname, int addPtr);

  protected void finalize () throws Throwable
  {
    super.finalize ();
    if (mgrPtr == 0) return;
    
    int retval = checkZeroRef ();
    while (checkZeroRef () != 0)
      Thread.currentThread ().sleep (100);
    
    if (retval != 0)
      log.warn (retval + " unexpected non-zero reference count");
    else
      log.debug ("All went well");
    quit ();
    mgrPtr = 0;
  }
  
  
  public int restrictVar (int ptr, int var, boolean val)
  {
    int varAdd = addVarPtr (var);
    
    if (!val)
      {
	int temp = addNotPtr (varAdd);
	recursiveDeref (varAdd);
	varAdd = temp;
      }

    int result = addConstrainPtr (ptr, varAdd);
    recursiveDeref (varAdd);
    return result;
  }
  
  
  // -- returns a add corresponding to 'index'
  public ADD addVar (int index)
  {
    return new ADD (this, addVarPtr (index));
  }
  
  public ADD addOne ()
  {
    if (addOne == null)
      addOne = new ADD (this, cuddAddOnePtr ());
    return addOne;
  }
  
  public ADD addZero ()
  {
    if (addZero == null)
      addZero = new ADD (this, cuddAddZeroPtr ());
    return addZero;
  }
  

  public String toString ()
  {
    return String.valueOf (mgrPtr);
  }
  
  public static void main (String[] args)
  {
    System.out.println ("Hello");
  }
  
//    public static void main (String[] args)
//    {
//      CuddLatAdd CuddLatAdd = new CuddLatAdd ();
//      System.out.println ("Got: " + CuddLatAdd);
//      System.out.println ("1: " + CuddLatAdd.addOnePtr ());
//      System.out.println ("0: " + CuddLatAdd.addZeroPtr ());
//      System.out.println ("1 * 0: " + 
//  			CuddLatAdd.addAndPtr 
//  			(CuddLatAdd.addOnePtr (), 
//  			 CuddLatAdd.addZeroPtr ()));
//      System.out.println ("1 + 0: " + 
//  			CuddLatAdd.addOrPtr 
//  			(CuddLatAdd.addOnePtr (), 
//  			 CuddLatAdd.addZeroPtr ()));
    
//    }

  public boolean equals (Object o)
  {
    return o != null && o instanceof CuddLatAdd && equals ((CuddLatAdd)o);    
  }

  public boolean equals (CuddLatAdd o)
  {
    return getMgrPtr () == o.getMgrPtr ();
  }
  
  int getMgrPtr ()
  {
    return mgrPtr;
  }
  
  

  public static class ADD
  {
    CuddLatAdd manager;
    int nodePtr;
    
    protected ADD (CuddLatAdd _manager, int _nodePtr)
    {
      manager = _manager;
      nodePtr = _nodePtr;
    }

    public ADD addAnd (ADD add)
    {
      return 
	new ADD (manager, manager.addAndPtr (nodePtr, add.getNodePtr ()));
    }
    public ADD addOr (ADD add)
    {
      return new ADD (manager, manager.addOrPtr (nodePtr, add.getNodePtr ()));
    }
    public ADD addNot ()
    {
      return new ADD (manager, manager.addNotPtr (nodePtr));
    }
    public ADD constrain (ADD c)
    {
      return 
	new ADD (manager, manager.addConstrainPtr (nodePtr, c.getNodePtr ()));
    }
    
    public String toString ()
    {
      return manager + "@" + nodePtr;
    }
    
    
    CuddLatAdd getManager ()
    {
      return manager;
    }
    
    int getNodePtr ()
    {
      return nodePtr;
    }
    
    protected void finalize () throws Throwable
    {
      super.finalize ();
      if (nodePtr != 0)
	manager.recursiveDeref (nodePtr);
    }
  

    public boolean equals (Object o)
    {
      return o != null && o instanceof ADD && equals ((ADD)o);
    }
    
    public boolean equals (ADD o)
    {
      return getManager ().equals (o.getManager ()) &&
	getNodePtr () == o.getNodePtr ();
    }
    
    
  }
  
	
  
}
