package edu.toronto.cs.experiments;

import edu.toronto.cs.algebra.*;
import edu.toronto.cs.mvset.*;
import edu.toronto.cs.util.*;
import edu.toronto.cs.modelchecker.*;

import java.io.*;


/****
 **** An attempt at a flexible experiment runner
 ****/
public class ExperimentRunner
{
  
  static String suffix = ".dat";

  /** the output should go here */
  PrintWriter out;

  /** function we are working with */
  MvFunction function;
  
  /** number of samples */
  int sampleSize;

  /** MvSetFactory we are using */
  MvSetFactory factory;

  /** should we perform reordering each time */
  boolean reorder;

  String davinciFileName;
  
  boolean daVinci = true;

  public ExperimentRunner (PrintWriter _out, MvFunction _function,
			   int _sampleSize, MvSetFactory _factory, 
			   boolean _reorder, 
			   String _daVinciFileName)
  {
    out = _out;
    function = _function;
    sampleSize = _sampleSize;
    factory = _factory;
    reorder = _reorder;
    davinciFileName = _daVinciFileName;
  }
  

  public void run ()
  {
    for (int i = 0; i < sampleSize; i++)
      {
	// -- generate a new function
	function = function.newMvFunction ();
	System.out.println ("The function is: " );
	System.out.println (function);
	
	// -- convert it into and MvSet
	MvSet mvSet = function.toMvSet (factory);

	if (reorder)
	  mvSet.reorder ();
	
	int size = mvSet.size ();
	
	writeSample (i, size);

	if (daVinci)
	  {
	    // -- we don't have more variables than the size of the diagram
	    String[] propNames = new String [size >> 1];
	    for (int j = 0; j < propNames.length; j++)
	      propNames [j] = String.valueOf (j);
	    
	    PropOrderIf ordering = new AlternatePropOrder (propNames);
	    if (mvSet instanceof MDDMvSet)
	      {
		MDDMvSet mddMvSet = (MDDMvSet)mvSet;
		mddMvSet.writeToDaVinci (davinciFileName + ".daVinci", 
					 ordering);
	      }
	    else if (mvSet instanceof JMvSet)
	      {
		JMvSet jMvSet = (JMvSet)mvSet;
		jMvSet.writeToDaVinci (davinciFileName, ordering);
	      }
	  }
	
      }

    out.close ();
    out = null;
    
  }
  
  
  void writeSample (int sampleCount, int sample)
  {
    /*** currently we just \t separate things */
    out.print (sample);
    out.print ('\t');
    out.flush ();
  }
  

  public static void main (String[] args) throws Exception
  {
    /*** Seed the random number generator **/
    //Chaos.seed (1000);
    
    /**** Arguments are as follows
     **** logic mvset number_of_variables number_of_samples 
     **** output_prefix do_reordering function_type function_arguments
     ****/

    // -- get the lattice
    FiniteLattice lattice = Lattices.getLatticeByName (args [0]);
    
    // -- find out how many variables we want
    int numOfVars = Integer.parseInt (args [2]);

    // -- create an appropriate factory
    MvSetFactory factory = createFactory (args [1], lattice, numOfVars);
    
    int sampleSize = Integer.parseInt (args [3]);
    
    String fileNamePrefix = args [4];

    PrintWriter out = getPrintWriter (fileNamePrefix, args [1]);

    boolean reorder = new Boolean (args [5]).booleanValue ();
    
    String functionType = args [6];
    
    /** everything except for the first 7 arguments go to the function */
    String[] funcArgs = new String [args.length - 7];
    System.arraycopy (args, 7, funcArgs, 0, funcArgs.length);
    

    MvFunction function = createFunction (functionType, funcArgs, 
					  lattice, numOfVars);
    
    ExperimentRunner exp = new ExperimentRunner (out, function, sampleSize, 
						 factory, reorder,
						 fileNamePrefix + "_" + 
						 args [1]);
    exp.run ();
  }
  


  static PrintWriter getPrintWriter (String fileNamePrefix, String mvSetType)
    throws IOException
  {
    return new PrintWriter 
      (new FileWriter (fileNamePrefix + "_" + mvSetType +  suffix));
  }
  
  static MvSetFactory createFactory (String type, FiniteLattice lattice, 
				     int arrity)
  {
    if (type.equals ("mdd"))
      return MDDMvSet.newMvSetFactory (lattice, arrity);
    else if (type.equals ("jmbtdd"))
      return JMvSet.newMvSetFactory (lattice, arrity);
    else if (type.equals ("jbdd"))
      return JBDDMvSet.newMvSetFactory (lattice, arrity);
    else if (type.equals ("jadd"))
      return JADDMvSet.newMvSetFactory (lattice, arrity);
    else if (type.equals ("cubdd"))
      return CUBDDMvSet.newMvSetFactory (lattice, arrity);
    else if (type.equals ("cuadd"))
      return CUADDMvSet.newMvSetFactory (lattice, arrity);
    else
      Assert.assert ("Unknown mvset type: " + type);

    // -- We never actually get here
    return null;
  }
  
  static MvFunction createFunction (String type, String[] args, 
				    FiniteLattice lattice, int arrity)
  {
    if (type.equals ("mc"))
      return createMcFunction (lattice, arrity, args);
    else
      Assert.assert ("Unknown function type: " + type);
    
    // -- we never get here
    return null;
  }
  
  static ModelCheckingFunction createMcFunction (FiniteLattice lattice, 
						 int arrity, 
						 String[] args)
  {
    
    /***
     *** args [0] -- supportLow
     *** args [1] -- supportHigh
     *** args [2] -- number of 'minterms'
     ***/
    try {
      int supportLow = Integer.parseInt (args [0]);
      int supportHigh = Integer.parseInt (args [1]);
      int size = Integer.parseInt (args [2]);
      
      System.out.println ("Creating a function: " + arrity + " " + 
			  supportLow + " " + supportHigh + " " + size);

      return 
	new ModelCheckingFunction 
	(lattice, arrity, supportLow, supportHigh, size);
    }
    catch (Exception ex)
      {
	ex.printStackTrace ();
	System.exit (0);
      }
    
    return null;
  }
  
}
