package edu.toronto.cs.keg;

import java.util.*;
import java.io.*;

import edu.toronto.cs.algebra.*;
import edu.toronto.cs.util.dgraph.*;
import edu.toronto.cs.ctl.*;
import edu.toronto.cs.mvset.*;
import edu.toronto.cs.heuristics.*;
import edu.toronto.cs.modelchecker.*;

public class KEEUiSimplifier implements KEGExpander
{
  public KEEUiSimplifier ()
  {
  }
  
  public boolean expand (DGNode state, DGNode formula, MvSetModelChecker mc)
  {
    KEGCtlNode ctlNode = (KEGCtlNode)formula;
    
    DGraph ctlGraph = ctlNode.getGraph ();
    
    MvState s0 = ((KEGStateNode) state).getState ();
    
    if (ctlNode.getCTL ().getClass () == CTLEqualsNode.class)
      {
	// -- break the formula on equality
	CTLNode ctl = ctlNode.getCTL ().getLeft ();
	CTLLatticeNode value = (CTLLatticeNode) ctlNode.getCTL ().getRight ();
	// -- check that the formula is EUi, if not we are not handling it
	if (ctl.getClass () != CTLEUiNode.class) return false;
	
	// -- check that the value of the formula is correct, if not
	// -- some one else should handle it
	if (mc.checkCTL (ctl).evaluate (s0) != value.getLatIdx ()) 
	  return false;
	
	// -- try to expand EUi
	CTLNode simpleCTL = 
	  expandEUi (s0, (CTLEUiNode)ctl, value.getLatIdx (), mc);
	
	// -- success, add the formula to the tree and 
	// -- terminate successfully, otherwise we can't do anything
	if (simpleCTL != null)
	  {
	    KegUtils.addCtl (state, ctlNode, simpleCTL.eq (value));
	    return true;
	  }
      }
    return false;
  }
  

  /***
   *** s0 -- initial state
   *** eui -- the EUi formula
   *** value -- value at which EUi should be evaluated
   *** mc -- model-checker
   ***/
  CTLNode expandEUi (MvState s0, CTLEUiNode eui, int value, 
		     MvSetModelChecker mc)
  {


    // -- if the current bound is 0 then just simplify
    // -- E [a U_0 b] <=> b
    if (eui.getI () == 0)
      return eui.getRight ();
    
    /***
     *** keep decreasing the bound until we get a different result, 
     *** at this point we've found the lowest possible bound, 
     *** and if it is new we can use it
     ***/
    CTLNode phi = eui.getLeft ();
    CTLNode psi = eui.getRight ();
    
    int newBound = eui.getI ();
    for (int i = eui.getI () - 1; i >= 0; i--)
      {
	int val = mc.checkCTL (phi.eu (i, psi)).evaluate (s0);
	if (val != value)
	  {
	    // -- this is the breaking point where there is no reason
	    // -- to go any further
	    newBound = i + 1;
	    break;
	  }
	
      }
    
    // -- check if we got a new bound
    if (newBound != eui.getI ())
      return phi.eu (newBound, psi);

    return null;
  }

  public HeuristicAnalyser getHeuristics ()
  {
    return null;
  }
  
  public void setHeuristics(HeuristicAnalyser h)
  {
    
  }
  

  
  
}
