package edu.toronto.cs.modelchecker;

import edu.toronto.cs.util.*;
import edu.toronto.cs.util.userprompt.*;
import edu.toronto.cs.util.dgraph.*;
import edu.toronto.cs.prettygraph.*;
import edu.toronto.cs.xkripke.*;
import edu.toronto.cs.xkripke.XKripke.*;
import edu.toronto.cs.algebra.*;
import edu.toronto.cs.mvset.*;
import edu.toronto.cs.mdd.*;
import edu.toronto.cs.ctl.*;
import edu.toronto.cs.smv.*;

import edu.toronto.cs.tlq.UpSetLattice;


import java.util.*;
import java.io.*;
import java.lang.reflect.Method;

/**
 ** This is the front end of the model-checker. It deals with the user
 ** supplied parameters and drives the model-checker.
 **/ 
public class ModelCheckerFrontEnd
{

  // initialize the logger
  static Logger log = new Logger (ModelCheckerFrontEnd.class.getName ());

  // the xkripke structure(s)
  XKripke[] xkripke;

  // the propositional ordering
  PropOrderIf propOrder;

  // the associated model-checker
  MvSetModelChecker mc;
  
  // various options
  Map options;

  // verbosity level: 0 - lowest, 2 - highest (deafult)
  int verbose;

  // initial state
  MvState initState;

  // the macro expander for handling state names
  //  StringMacroManager macro;
  CTLMacroManager macroManager;

  PlaceholderManager2 placeHolderManager = null;

  // -- enumerated types that we know about
  EnumeratedTypes enumTypes = EnumeratedTypes.EMPTY_TYPES;

  /**
   ** Initialises the modelchecker front end with the specified options.
   ** 
   ** @param opts -- a Map of various options
   **/
  public ModelCheckerFrontEnd (Map opts)
  {
    // parse the command line options
    options = opts;

    log.info ("Options passed to the modelchecker: " + options);

    // get the verbosity level: 0 - lowest, 2 - highest (default)
    verbose = ((Integer) options.get ("verbose")).intValue ();
    if (verbose < 0)
      verbose = 0;
    else if (verbose > 2)
      verbose = 2;

    MvSet transRelation = null;
    
    // XXX This was not thought through, but it will WORK! hopefully
    
    if (((Boolean)options.get ("smv")).booleanValue ())
      {
	try {
	  transRelation = initMCUsingSMVModule ();
	} catch (Exception ex) {
	  ex.printStackTrace ();
	  throw new IllegalArgumentException ("Some error or other");
	}

      }
    // B.D. : kludge to use DDUnDumper to load in a transition system.
    else if (((Boolean)options.get("bdd")).booleanValue()) {
      
      System.out.println ("You choose poorly.");
      
      // get the BDD file name
      // get the variable ordering file name
      // load the BDD file, make a transition relation
      // load the variable-ordering file, make a propositional ordering
	
      }
    else
      {
	// generate the xkripke structure from the input file(s)
	makeXKripke ();
	
	// -- the following components are needed to initialize 
	// -- the modelchecker
	// create the propositional ordering
	makePropOrder ();
	//    propOrder = new AlternatePropOrder (xkripke.getPropNames ());
	
	// get the MvSet representation of the transition relation
	transRelation = createTransRelation ();
	
	// get the initial state
	initState = getInitialState ();
	//System.out.println ("The initial state is: " + initState);
      
      }
    // create the model-checker
    mc = new MvSetModelChecker (transRelation,
				propOrder);
    // -- createa a macro manager
    macroManager = new CTLMacroManager (mc.getLattice ());

    
    // -- if possible create a place-holder manager
    if (mc.getLattice () instanceof UpSetLattice)
      {
	placeHolderManager = 
	  new PlaceholderManager2 ((UpSetLattice)mc.getLattice (), enumTypes);
	macroManager.addExpander (placeHolderManager);
      }

    // -- we have enumerated types, so create macroExpander for them
    if (enumTypes != null && enumTypes != EnumeratedTypes.EMPTY_TYPES)
      macroManager.addExpander (new SMVEnumTypeExpander (enumTypes));
    
    
	
    System.out.println ("Some statistics: ");
    System.out.println ("Size of trans: " + transRelation.size ());
//     System.out.println ("Saving transition relation to a file");
//     transRelation.writeToDaVinci ("/tmp/transrel.daVinci", propOrder);

    //     transRelation = mc.computeRechable (initState);
     //     System.out.println ("Size of reachable trans: " + transRelation.size ());
//     if (transRelation.size () < mc.getMvTransRelation ().size ())
//       mc.setTransitionRelation (transRelation);


    
//     transRelation.reorder ();
//     System.out.println ("Size of reachable trans after reorder: " +
//  			transRelation.size ());

  }

  /**
   ** The driver for the model-checker.
   **/
  public static void main (String[] args) throws Exception // XXX -- V: fix it
  {

    Map opts = McClapi.parseOpts (args);
    ModelCheckerFrontEnd frontend = new ModelCheckerFrontEnd (opts);
    
    frontend.runMC ();
  }

  /**
   ** Asks the user for CTL formulas and modelchecks them.
   **/
  public void runMC () throws Exception // XXX -- Victor: need to fix it
  {
    NiceReader input = new NiceReader (System.in);
    // -- tell the converter what lattice it should use
    BinaryTreeToCTLConverter.AssignLattice (mc.getLattice ());

    boolean run = true;
    do
      {
	try {
	  run = checkOneProp (input);
	} catch (Exception e) {
	  System.out.println ("Error while model checking: " + e);
	  e.printStackTrace ();
	  // -- don't quit if we have a problem, just move on
	  run = true;
	}
	
      }
    while (run);
  }


  /** 
   * Reads a single property from standard input,
   * runs the model checker,
   * and outputs the result. This functions is called continuously from 
   * the main loop
   **/
  boolean checkOneProp (NiceReader input)
  {
    // stuff needed to read the CTL formulas
    String ctlStr;
    CTLNode ctl;
	
    
    // -- reset macro manager each time
    macroManager.reset ();

    // print the propositional ordering so that it is easier to
    // construct a CTL
    //printPropOrdering ();

    // read a CTL
    System.out.println 
      ("What CTL property would you like to check today?");
    ctlStr = input.readLine ();
    if (ctlStr.length () == 0) return false;// empty input - exit the loop
    else if (ctlStr.equals ("state"))
      {
	System.out.println ("Type in the name of the new initial state" +
			    "(leave blank to keep the old one):");
	String name = input.readLine ();
	options.put ("initstate", name);
	initState = getInitialState ();
      }
    else if (ctlStr.equals ("macro"))
      {
	System.out.println ("Type in the name of the new macro:");
	String name = input.readLine ();
	System.out.println ("Type in the expansion of the new macro:");
	String expansion = input.readLine ();
	// -- transform into the standard CTL String representation
	CTLNode exp = BinaryTreeToCTLConverter.convertToCTL (expansion);
	    
	macroManager.defineMacro (CTLFactory.createCTLAtomNode (name),
				  exp);
      }
    else if (ctlStr.equals ("macro_reset"))
      {
	System.out.println ("Reseting macros");
	macroManager.reset ();
      }
    
    else
      {
	if (verbose > 0) // XXX -- V: HACK! change to constants soon . . .
	  System.out.println ("Parsing: " + ctlStr);    

	// -- convert the ctlStr into a CTL tree
	ctl = BinaryTreeToCTLConverter.convertToCTL (ctlStr);
	// -- expand the macroes -- called twice for 2 pass expansion
	ctl = macroManager.expandMacro (ctl);
	ctl = macroManager.expandMacro (ctl);

	//	    ctlStr = macro.expandMacro (ctlStr);
	if (verbose > 0) // XXX -- V: HACK! change to constants soon . . .
	  System.out.println ("After macro expansion: " + ctl);    

	// -- run our model checker on the property
	// -- record the time that it takes
	StopWatch stopwatch = new StopWatch ();
	MvSet mvResult = mc.doCTL (ctl);
	stopwatch.stop ();
	if (verbose > 0) // this will do for now
	  {
	    System.out.println ("time: " +
				stopwatch.getTimeElapsed ()/1000.0 + "s");
	    System.out.println ("size of trans: " + 
				mc.getMvTransRelation ().size ());
	    System.out.println (mvResult.getFactory ().toString ());
	  }
	    


	// -- print the result
	if (verbose > 0) // XXX -- V: HACK! change to constants soon . . .
	  System.out.println
	    ("This is the formula that you are checking: " +
	     macroManager.compressMacro (ctl));

	// 	    if (mvResult instanceof MDDMvSet)
	// 	      writeMDDToFile ((MDDMvSet) mvResult);

	// 	    System.out.println ("Final result would've been here..." + 
	// 				" disabled for now");
	// XXX
	printResult (mvResult);
	System.out.println ("Writing result to a file");
	mvResult.writeToDaVinci ("/tmp/result.daVinci", propOrder);

      }
    return true;
  }
  

  /**
   ** Prints the result of the model-checking restricted to the
   ** initial state.
   **/
  public void printResult (MvSet mvResult)
  {
    // -- restrict the result to an initial state, this should result 
    // -- in a constant MDD, i.e. a simple T, F, etc. answer
    if (verbose > 0)
      System.out.println ("Restricting to initial state");
    mvResult = restrictToInitialState (mvResult);

    

    // get the lattice value of the result
    int resultValue = mvResult.evaluate 
      (new MvState (new int [0], mc.getLattice ()));

    System.out.println ("The result is: " + resultValue);

    // -- if we have a place holder manager, use it to interpret the 
    // -- solution
    if (placeHolderManager == null)
      {
	System.out.println ("Final result: " + 
			    mc.getLattice ().
			    getElementKeyByIdx (resultValue));
      }
    else
      {
	System.out.println ("Final result:");
	List solution = placeHolderManager.interpretSoln (resultValue);
	int count = 0;
	for (Iterator it = solution.iterator (); it.hasNext ();)
	  {
	    System.out.println ("Solution " + (count++) + ": ");
	    System.out.println (it.next ());
	  }
	
      }
    
    
    // depending on the verbosity print additional info
    PrintWriter out = new PrintWriter (System.out);
    switch (verbose)
      {
      case 2:
	// -- print a map that maps user's idea about logic values and 
	// -- internal representation used in debug print outs
	mc.getLattice ().dumpValueMap (out);
      case 1: // nothing for now
      case 0: break;
      }
    out.flush ();
  }

  /**
   ** Prints the current propositional ordering.
   **/
  public void printPropOrdering ()
    {
      if (verbose > 0)
	{
	  int size = propOrder.size ();
	  for (int i = 0; i < size; i++)
	    {
	      if (propOrder.isPostImage (i))
		{
		  if (verbose > 1)
		    System.out.println
		      ("#" + i + " = " + propOrder.getPropName (i) + "'");
		}
	      else
		System.out.println
		  ("#" + i + " = " + propOrder.getPropName (i));
	    }
	}
    }

  /**
   ** Writes the MDD to a file in daVinci format.
   **/
  public void writeMDDToFile (MDDMvSet mddmvset) throws Exception
  {
    // get the MDD from the MvSet
    MDDNode result = mddmvset.getMddNode ();
    // get the output filename
    String filename = (String) options.get ("output");

    if (result != null)
      {
	// -- write a graphical representation of resulting mdd
	System.out.println ("AM SO writing mdd to " + filename);

	// Below should work, but does (or at least 1) creates output
	// and 2) doesn't crash).
	DGraph tmp = PrettyGraphMDDVisitor.getPrettyRep 
	  (result, propOrder);
	DVGenNodeVisitor dvg = new DVGenNodeVisitor 
	  (new DVGenEdgeVisitor ());
	for (Iterator it = DGDepthTraverse.getIterator (tmp);
	     it.hasNext ();)
	  {
	    dvg.visit ((DGNode)it.next (), null);
	  }
	StringBuffer sb = dvg.getSB ();
	System.out.println ("And here it is:"+sb);
	// output the daVinci graph to the file specified in
	// the options (default mdd.daVinci)
	FileWriter dvmdd = new FileWriter (filename);
	dvmdd.write (sb.toString ());
	dvmdd.close ();
      }
  }
  
  /**
   ** Creates the appropriate propositional ordering for all the
   ** XKripke structires.
   **/
  private void makePropOrder ()
  {
    // first get all the propositional variables in one set
    List allprops = new ArrayList ();
    
    for (int x = 0; x < xkripke.length; x++)
      allprops.addAll (xkripke [x].getPropNamesAsSet ());
    // now feed them to the AlternatePropOrder constructor
    propOrder = new AlternatePropOrder ((String []) allprops.toArray 
					(new String [allprops.size ()]));
  }

  /**
   ** Generates the xkripke structure(s) from the input file(s) specified
   ** in the options.
   **/
  private void makeXKripke ()
    throws IllegalArgumentException, NestedRuntimeException
  {
    // get the file name(s)
    List filenames = (List) options.get ("input");

    // determine whether states should be labelled
    boolean addStateNames = ((Boolean) options.get ("statenames")).
      booleanValue ();

    if (filenames.size () == 0)
      {
	throw new IllegalArgumentException 
	  ("Cannot create XKripke structure: no input file specified.");
      }
    else
      {
	try
	  {
	    xkripke = XKripkeFactory.parse 
	      ((String []) filenames.toArray ());
	  }
	catch (Exception e)
	  {
	    throw new NestedRuntimeException (e);
	  }
      }

    // add state name labels if needed
    if (addStateNames)
      for (int x = 0; x < xkripke.length; x++) xkripke [x].labeliseStates ();
  }

  /**
   ** Constructs the transition relation represented as an MvSet from
   ** the xkripke structure.
   **
   ** @return -- the transition relation corresponding to the
   ** specified xkripke structure
   **/
  private MvSet createTransRelation ()
  {
    // get the lattice -- assumes that all the lattices used are the same
    FiniteLattice lattice = xkripke [0].getLattice ();

    MvSetFactory mvSetFactory = createMvSetFactory (lattice);

    // construct the MvSet using the appropriate factory
    return XKripkeMvSetBuilder.buildMvSet (mvSetFactory, xkripke,
					    propOrder);
  }

  /**
   ** Creates an MvSetFactory according to the options.
   **/
  private MvSetFactory createMvSetFactory (FiniteLattice lattice)
  {
    // get the MvSet type
    String mvsettype = (String) options.get ("mvset");

    // select MvSetFactory
    if (mvsettype.equals ("mdd")) // -- MDD based
      {
	log.debug ("Using MDDMvSets");
	return MDDMvSet.newMvSetFactory (lattice, propOrder.size ());
      }
    else if (mvsettype.equals ("join")) // -- Join Irreducible based
      {
	log.debug ("Using JMvSets");
	return JMvSet.newMvSetFactory (lattice, propOrder.size ());
      }
    else if (mvsettype.equals ("jbdd")) // -- JBDD implementation
      {
	log.debug ("Using JBDDMvSets");
	return JBDDMvSet.newMvSetFactory (lattice, propOrder.size ());
      }   
    else if (mvsettype.equals ("cubdd"))
      {
	log.debug ("Using CUBDDMvSet");
	return CUBDDMvSet.newMvSetFactory (lattice, propOrder.size ());
      }
    else if (mvsettype.equals ("cuadd"))
      {
	log.debug ("Using CUADDMvSet");
	return CUADDMvSet.newMvSetFactory (lattice, propOrder.size ());
      }
    else if (mvsettype.equals ("jadd"))
      {
	log.debug ("Using JADDMvSet");
	return JADDMvSet.newMvSetFactory (lattice, propOrder.size ());
      }
    else if (mvsettype.equals ("cubiadd"))
      {
	log.debug ("Using CUBiADDMvSet");
	
	return CUBiADDMvSet.newMvSetFactory (lattice, propOrder.size ());
      }
    else throw new IllegalArgumentException 
	   ("MvSet type '" + mvsettype + "' not supported."); 
  }

  /**
   ** Gets the initial state for the current model.
   **
   ** @return -- the initial state for the model or null if no unique
   ** initial state was defined.
   **/
  private MvState getInitialState ()
  {
    // get the name of the initial state from the options
    String name = (String) options.get ("initstate");
    FiniteLattice lattice = xkripke [0].getLattice ();
    MvState initstate = new MvState (propOrder.size (), lattice);
    MvState [] states = new MvState [xkripke.length];

    if (name == null) // no initial state was specified - use default
      {
	// unify all the initial states
	for (int x = 0; x < xkripke.length; x++)
	  {
	    // get the initial state of the current model
	    states [x] = new MvState (xkripke [x].getInitialStates () [0],
				      propOrder);
	  }
      }
    else // a particular state was selected
      {
	StringTokenizer namelist = new StringTokenizer (name, ",");

	for (int x = 0; x < xkripke.length; x++)
	  states [x] = new MvState (xkripke [x].getState
				      (namelist.nextToken ()), propOrder);
      }

    // copy all the valid values of the propositions
    for (int x = 0; x < xkripke.length; x++)
      for (int i = 0; i < initstate.size (); i++)
	if (states [x].getValue (i) != Algebra.NO_VALUE)
	  initstate.setValue (i, states [x].getValue (i));

    return initstate;
  }

  /**
   ** Restricts the result to a unique initial state if one exists,
   ** otherwise does nothing.
   **/
  private MvSet restrictToInitialState (MvSet result)
    {
      // -- restrict the result to a unique initial state if one exists
      if (initState == null)
	return result; // do nothing
      else // restrict the result to initState
	{
// 	  System.out.println ("Restricting with respect to: ");
// 	  System.out.println (initState.toString (propOrder));
	  return restrictMvSet (initState, result);
	}
    }

  /**
   ** Given an MvState restricts the MvSet to it in order to obtain a
   ** single answer.
   **/  
  private MvSet restrictMvSet (MvState state, MvSet mvSet)
    {
      return mvSet.restrict (state);
      //return result;
    }  

  /**
   ** Gets the model-checker.
   **/
  MvSetModelChecker getMC ()
  {
    return mc;
  }

  // -- XXX
  public MvSet initMCUsingSMVModule () throws Exception
  {
    // -- here we initialize the model checker using SMV module from 
    // -- command line argument. 
    // -- we need to initialize
    // -- 1) Lattice
    // -- 2) Prop ordering
    // -- 3) Transition relation
    // -- 4) Initial state
    
    List smvModelCmdLine = (List)options.get ("input");
    
    // -- extract smvModule name and its arguments
    String smvModuleName = (String)smvModelCmdLine.get (0);
    String[] smvModuleArgs = 
      (String[])smvModelCmdLine.subList (1, smvModelCmdLine.size ()).
      toArray (new String [smvModelCmdLine.size () - 1]);
    
    SMVModule smvModule = loadSMVModule (smvModuleName, smvModuleArgs);

    FiniteLattice lattice = smvModule.getLattice ();

    
    propOrder = new AlternatePropOrder (smvModule.getVars ());
    
    smvModule.setVarOrdering (propOrder);
    System.out.println("varsLength is now: "+smvModule.getVars().length);
    MvSetFactory ddManager = createMvSetFactory (lattice);
    smvModule.setMvSetFactory (ddManager);

    // -- set init state
    initState = new MvState (propOrder.size (), lattice);
    initState = smvModule.initializeInitState (initState);
    

    // -- enumerated types side-effect
    enumTypes = smvModule.getEnumTypes ();

    // -- compute transition relation
    return smvModule.computeTrans ();
  }

  
  private SMVModule loadSMVModule (String className, String[] args) 
    throws Exception
  {
    Class clazz = Class.forName (className);
    Method initMethod = clazz.getMethod ("initModule", 
					 new Class[]{String[].class});
    return (SMVModule)initMethod.invoke (clazz, new Object[]{args});
  }

}


